package flux;

import java.io.Serializable;

/**
 * Git trigger. Watches remote git repository and fires when a new revision is seen.
 * <p/>
 * <h3>Usage Example</h3>
 * <p/>
 * Here is a sample job that watches Jersey Git repository every hour and fires when a
 * new changeset is seen:
 * <p/>
 * <pre> {@code
 * FlowChart flowChart = Factory.makeInstance().makeEngineHelper().makeFlowChart("Git Trigger Example");
 * GitTriggerFactory triggerFactory = (GitTriggerFactory) flowChart.makeFactory("GitTriggerFactory");
 * GitTrigger gitTrigger = triggerFactory.makeGitTrigger("Git");
 * gitTrigger.setRepositoryUrl("https://github.com/jersey/jersey.git");
 * gitTrigger.setPollingDelay("+1H");
 * }</pre>
 *
 * @author arul@flux.ly
 */

public interface GitTrigger extends Trigger {

  /**
   * Sets the path of the hg executable.
   *
   * @param path the path of the hg executable.
   */
  void setPath(String path);

  /**
   * Sets the delay that occurs between changelog polling. Defaults to +1m, one minute.
   *
   * @param timeExpression the delay that occurs between changelog polling.
   */
  public void setPollingDelay(String timeExpression);

  /**
   * Sets the remote git repository URL.
   *
   * @param url the remote git repository URL.
   */
  void setRepositoryUrl(String url);

  /**
   * The response from the GitTrigger.
   */
  public class GitTriggerResult implements Serializable, Cloneable {
    /**
     * Latest commit id.
     */
    public String commitId;

    public Object clone() throws CloneNotSupportedException {
      return super.clone();
    } // clone()
  }
}

