package flux;

import java.io.Serializable;

/**
 * Mercurial trigger. Watches remote mercurial repository and fires when a new revision is seen.
 * <p/>
 * <h3>Usage Example</h3>
 * <p/>
 * Here is a sample job that watches OpenJDK Mercurial repository every hour and fires when a
 * new changeset is seen:
 * <p/>
 * <pre> {@code
 * FlowChart flowChart = Factory.makeInstance().makeEngineHelper().makeFlowChart("Mercurial Trigger Example");
 * MercurialTriggerFactory triggerFactory = (MercurialTriggerFactory) flowChart.makeFactory("MercurialTriggerFactory");
 * MercurialTrigger mercurialTrigger = triggerFactory.makeMercurialTrigger("HgT");
 * mercurialTrigger.setRepositoryUrl("http://hg.openjdk.java.net/jdk7/jdk7");
 * mercurialTrigger.setPollingDelay("+1H");
 * }</pre>
 *
 * @author arul@flux.ly
 */
public interface MercurialTrigger extends Trigger {

  /**
   * Sets the path of the hg executable.
   *
   * @param path the path of the hg executable.
   */
  void setPath(String path);

  /**
   * Sets the delay that occurs between changelog polling. Defaults to +1m, one minute.
   *
   * @param timeExpression the delay that occurs between changelog polling.
   */
  public void setPollingDelay(String timeExpression);

  /**
   * Sets the remote mercurial repository URL.
   *
   * @param url the remote mercurial repository URL.
   */
  void setRepositoryUrl(String url);

  /**
   * The response from the MercurialTrigger.
   */
  public class MercurialTriggerResult implements Serializable, Cloneable {
    /**
     * Latest change set id.
     */
    public String changeSet;

    public Object clone() throws CloneNotSupportedException {
      return super.clone();
    } // clone()
  }
}
