package flux;

import java.io.Serializable;

/**
 * SCPAction that transfers files over an SSH connection or runs a remote
 * command. If remote command is set, it takes precedence over transfer.
 *
 * If you don't already have ~/.ssh/id_rsa present, generate a key with the
 * command 'ssh-keygen -t rsa -P '''. The null private key password allows
 * for automated SSH connections.
 *
 * @author arul@flux.ly
 */
public interface SCPAction extends Action {

  /**
   * Set the compression flag. Defaults to false. Uses JZlib for compression.
   *
   * @param compression compression.
   */
  public void setCompression(boolean compression);

  /**
   * Set the download flag. Defaults to true. For, upload set this to false.
   *
   * @param download download.
   */
  public void setDownload(boolean download);

  /**
   * Set the Host.
   *
   * @param host name of the host where broker is running.
   */
  public void setHost(String host);

  /**
   * Set the local path.
   *
   * @param localPath Local path.
   */
  public void setLocalPath(String localPath);

  /**
   * Set the password used to authenticate username using the "password"
   * authentication method and as a fallback basic challenge-response
   * authentication.
   *
   * @param password password.
   */
  public void setPassword(String password);

  /**
   * Set the fully qualified private key file name. If not set, it would
   * look for private keys ~/.ssh/id_rsa and ~/.ssh/id_dsa. If none exists,
   * an IllegalArgumentException is thrown.
   *
   * @param privateKey private key.
   *
   * @throws IllegalArgumentException If private key is null or not available.
   */
  public void setPrivateKey(String privateKey);

  /**
   * Set the remote command.
   *
   * @param command command.
   */
  public void setRemoteCommand(String command);

  /**
   * Set the remote path.
   *
   * @param remotePath Remote path.
   */
  public void setRemotePath(String remotePath);

  /**
   * Set the username.
   *
   * @param username username.
   */
  public void setUsername(String username);

  /**
   * The response from SCPAction that executes a remote command.
   */
  public class SCPActionResult implements Serializable, Cloneable {

    /**
     * The process's exit code, as returned by the operating system.
     */
    public int result;

    /**
     * The standard error (stderr) from the completed process. This field is
     * transient, so its value, if any, will not be stored to the database.
     */
    public transient String stderr;

    /**
     * The standard output (stdout) from the completed process. This field is
     * transient, so its value, if any, will not be stored to the database.
     */
    public transient String stdout;

    @Override
    protected Object clone() throws CloneNotSupportedException {
      return super.clone();
    }
  }
}
