package flux;

import java.io.Serializable;
import java.util.Map;

/**
 * WebServiceTrigger that waits for a SOAP or HTTP request and starts a workflow.
 * These requests can be sent from any downstream systems. For example, in a loan
 * originations system, when a loan is accepted, the sales frontend can send a request
 * to a decision workflow that performs credit checks for the customer.
 * <p/>
 * The web service exposes both SOAP and REST endpoints.
 * <p/>
 * The SOAP WSDL is accessible at http://localhost:port/soap/SERVICE_NAME?wsdl
 * <p/>
 * The REST WADL is accessible at http://localhost:port/rest/SERVICE_NAME?_wadl&_type=xml
 *
 * @author arul@flux.ly
 */
public interface WebServiceTrigger extends Trigger {

  /**
   * Sets the password used to login to a remote Flux engine
   * to perform secured operations.
   *
   * @param enginePassword The password used to login to a
   * remote Flux engine to perform secured operations.
   */
  public void setEnginePassword(String enginePassword);

  /**
   * Sets the username used to login to a remote Flux engine
   * to perform secured operations.
   *
   * @param engineUsername The username used to login to a
   * remote Flux engine to perform secured operations.
   */
  public void setEngineUsername(String engineUsername);
  /**
   * Sets the password used to secure the web service.
   *
   * @param password The password used to secure the web service.
   */
  public void setPassword(String password);

  /**
   * Sets the port the web service listens to.
   *
   * @param port Port the web service listens to.
   */
  public void setPort(int port);

  /**
   * Sets the service name. Any spaces would get removed
   * as this name forms part of the URI.
   *
   * @param serviceName The name of the service.
   */
  public void setServiceName(String serviceName);

  /**
   * Sets the username used to secure the web service.
   *
   * @param username The username used to secure the web service.
   */
  public void setUsername(String username);

  /**
   * The response from the WebServiceTrigger.
   */
  public class WebServiceTriggerResult implements Serializable, Cloneable {

    /**
     * The result when a web service trigger fires. Indicates whether a web service
     * trigger is expedited automatically or manually. If true, the web service
     * trigger was expedited automatically. Otherwise, the web service trigger
     * was manually expedited.
     */
    public boolean expedited;

    /**
     * Flow chart variables set when expedite is called on this trigger.
     */
    public Map<String, String> variables;

    public Object clone() throws CloneNotSupportedException {
      return super.clone();
    } // clone()
  }
}