package fluximpl;

import flux.*;
import flux.dev.NotTriggeredException;

import javax.jms.*;
import javax.naming.Context;
import javax.naming.InitialContext;
import java.util.*;

/**
 * JmsTrigger which wait's for and consumes messages from a queue.
 *
 * @author Copyright 2012 Flux Corporation. All rights reserved.
 */
public class JmsTriggerImpl extends TriggerImpl implements JmsTrigger {

  private static final String JMS_TRIGGER_VARIABLE = "JMS_TRIGGER";

  /**
   * Timeout for {@link javax.jms.MessageConsumer#receive(long)}.
   */
  public static final long RECEIVE_TIMEOUT_INTERVAL = 1000L;


  // constructors
  public JmsTriggerImpl() {
    this(new FlowChartImpl(), "JMS Trigger");
  }

  /**
   * Constructor for end users. Ensures that every JavaAction belongs to a flow chart and has a name.
   *
   * @param fc   the flow chart the action belongs to.
   * @param name the name of the flow chart.
   */
  public JmsTriggerImpl(final FlowChartImpl fc, final String name) {
    super(fc, name);
  }

  public String getListeningQueueName() {
    return getVariable().listeningQueueName;
  }

  public void setListeningQueueName(String queueName) {
    JmsTriggerVariable var = getVariable();
    var.listeningQueueName = queueName;
    putVariable(var);
  }

  public boolean isFetchAllMessages() {
    return getVariable().fetchAllMessages;
  }

  public void setFetchAllMessages(boolean fetchAll) {
    JmsTriggerVariable var = getVariable();
    var.fetchAllMessages = fetchAll;
    putVariable(var);
  }

  public void setConnectionFactory(String connectionFactory) {
    JmsTriggerVariable var = getVariable();
    var.connectionFactory = connectionFactory;
    putVariable(var);
  }

  public String getConnectionFactory() {
    return getVariable().connectionFactory;
  }

  public String getInitialContextFactory() {
    return getVariable().initialContextFactory;
  }

  public void setInitialContextFactory(String initialContextFactory) {
    JmsTriggerVariable var = getVariable();
    var.initialContextFactory = initialContextFactory;
    putVariable(var);
  }

  public String getProviderUrl() {
    return getVariable().providerUrl;
  }

  public void setProviderUrl(String providerUrl) {
    JmsTriggerVariable var = getVariable();
    var.providerUrl = providerUrl;
    putVariable(var);
  }

  public String getUsername() {
    return getVariable().username;
  }

  public void setUsername(String username) {
    JmsTriggerVariable var = getVariable();
    var.username = username;
    putVariable(var);
  }

  public String getPassword() {
    Password password = getVariable().password;
    if (password != null) {
      return password.getEncryptedPassword();
    }
    return null;
  }

  public void setPassword(String password) {
    JmsTriggerVariable var = getVariable();
    if (password != null) {
      var.password = Password.makePassword(password);
    }
    putVariable(var);
  }

  public String getMessageSelector() {
    return getVariable().messageSelector;
  }

  public void setMessageSelector(String messageSelector) {
    JmsTriggerVariable var = getVariable();
    var.messageSelector = messageSelector;
    putVariable(var);
  }

  public String getPollingDelay() {
    String pollingDelay = getVariable().pollingDelay;
    if (pollingDelay == null) {
      pollingDelay = "+3s";
      setPollingDelay(pollingDelay);
    }
    return pollingDelay;
  }

  public void setPollingDelay(String pollingDelay) {
    JmsTriggerVariable var = getVariable();
    var.pollingDelay = pollingDelay;
    putVariable(var);
  }

  @Override
  public Properties getExtraInitialContextProperties() {
    return getVariable().extraInitialContextProperties;
  }

  @Override
  public void setExtraInitialContextProperties(Properties properties) {
    JmsTriggerVariable var = getVariable();
    var.extraInitialContextProperties = properties;
    putVariable(var);
  }

  public Object execute(FlowContext flowContext) throws Exception {
    String initialContextFactory = getInitialContextFactory();
    String connectionFactory = getConnectionFactory();
    String providerUrl = getProviderUrl();
    String jmsTriggerQueue = getListeningQueueName();

    Properties props = new Properties();
    props.setProperty("java.naming.factory.initial", initialContextFactory);
    props.setProperty("java.naming.provider.url", providerUrl);
    if (!checkNullOrEmptyString(getUsername())) {
      props.setProperty("java.naming.security.principal", getUsername());
      if (!checkNullOrEmptyString(getPassword())) {
        props.setProperty("java.naming.security.credentials", getPassword());
      }
    }
    if (getExtraInitialContextProperties().size() > 0) {
      props.putAll(getExtraInitialContextProperties());
    }

    Context ctx = new InitialContext(props);
    QueueConnectionFactory factory = (QueueConnectionFactory) ctx.lookup(connectionFactory);
    QueueConnection connection = factory.createQueueConnection();
    connection.start();

    //http://docs.oracle.com/javaee/1.4/api/javax/jms/Session.html#AUTO_ACKNOWLEDGE
    QueueSession session = connection.createQueueSession(false, 1);
    Destination destination = session.createQueue(jmsTriggerQueue);

    MessageConsumer consumer;
    if (!checkNullOrEmptyString(getMessageSelector())) {
      consumer = session.createConsumer(destination, getMessageSelector());
    } else {
      consumer = session.createConsumer(destination);
    }

    List<Message> messageList = null;
    if (isFetchAllMessages()) {
      QueueBrowser browser;
      if (!checkNullOrEmptyString(getMessageSelector())) {
        browser = session.createBrowser(session.createQueue(jmsTriggerQueue), getMessageSelector());
      } else {
        browser = session.createBrowser(session.createQueue(jmsTriggerQueue));
      }
      Enumeration e = browser.getEnumeration();
      int cnt = 0;
      while (e.hasMoreElements()) {
        cnt++;
        e.nextElement();
      }
      messageList = new ArrayList<Message>(cnt);
      for (int i = 0; i < cnt; i++) {
        final Message msg = consumer.receive(RECEIVE_TIMEOUT_INTERVAL);
        if ((msg != null)) {
          String correlationId = msg.getJMSCorrelationID();
          flowContext.getLogger().info("Message received, correlation ID = " + correlationId);
          messageList.add(msg);
        }
      }
    } else {
      Message msg = consumer.receive(RECEIVE_TIMEOUT_INTERVAL);
      if (msg != null) {
        messageList = new ArrayList<Message>(1);

        String correlationId = msg.getJMSCorrelationID();
        flowContext.getLogger().info("Message received, correlation ID = " + correlationId);
        messageList.add(msg);
      }
    }
    consumer.close();
    session.close();
    connection.close();

    final JmsTriggerResult result = prepareResult(messageList, flowContext);
    if (result != null) {
      flowContext.getLogger().info("Exiting JmsTrigger - " + result.size() + " messages RECEIVED");
    } else {
      throw new NotTriggeredException();
    }

    return result;
  }

  private JmsTriggerResult prepareResult(final List<Message> msgs, FlowContext flowContext) {
    if (msgs == null) {
      return new JmsTriggerResult();
    }
    List<Message> messageList = new ArrayList<Message>(msgs.size());
    final JmsTriggerResult result = new JmsTriggerResult(messageList);

    for (final Message message : msgs) {
      try {
        final Message emMessage = TextMessageImpl.makeActiveMQTextMessage(message);
        result.addMessage(emMessage);
      } catch (final JMSException ex) {
        flowContext.getLogger().info("Error converting message to POJO # " + ex.getMessage());
      }
    }

    return result;
  }


  public Date getNextPollingDate() {
    try {
      Factory fluxFactory = Factory.makeInstance();
      EngineHelper engineHelper = fluxFactory.makeEngineHelper();
      return engineHelper.applyTimeExpression(getPollingDelay(), null, null);
    } // try
    catch (EngineException e) {
      // Should not occur.
    } // catch
    return null;

  }

  public void verify() throws EngineException {
    if (checkNullOrEmptyString(getListeningQueueName())) {
      throw new EngineException("Expected \"Listening Queue Name\" to be non-null or non-empty, but it was null or empty.");
    }
    if (checkNullOrEmptyString(getConnectionFactory())) {
      throw new EngineException("Expected \"Connection Factory\" to be non-null or non-empty, but it was null or empty.");
    }
    if (checkNullOrEmptyString(getProviderUrl())) {
      throw new EngineException("Expected \"Provider URL\" to be non-null or non-empty, but it was null or empty.");
    }
    if (checkNullOrEmptyString(getInitialContextFactory())) {
      throw new EngineException("Expected \"Initial Context Factory\" to be non-null or non-empty, but it was null or empty.");
    }
    if ((checkNullOrEmptyString(getUsername())) && (!checkNullOrEmptyString(getPassword()))) {
      throw new EngineException("\"Password\" is set but \"Username\" is not. Both properties must be set for password-based authentication.");
    }
  }

  private JmsTriggerVariable getVariable() {
    if (!getVariableManager().contains("JMS_TRIGGER")) {
      getVariableManager().put("JMS_TRIGGER", new JmsTriggerVariable());
    }

    return (JmsTriggerVariable) getVariableManager().get("JMS_TRIGGER");
  }

  private void putVariable(JmsTriggerVariable variable) {
    getVariableManager().put("JMS_TRIGGER", variable);
  }

  @Override
  public Set<String> getHiddenVariableNames() {
    // The returned raw type is a Set of String.
    @SuppressWarnings("unchecked")
    Set<String> set = super.getHiddenVariableNames();
    set.add(JMS_TRIGGER_VARIABLE);
    return set;
  }

  private boolean checkNullOrEmptyString(String s) {
    return (s == null) || (s.equals(""));
  }
}
