package fluximpl;

import flux.EngineException;
import flux.Factory;
import flux.FlowContext;
import flux.SCPAction;
import fluximpl.audittrail.AuditTrailConstants;
import fluximpl.xml.five_zero_converter.ElementNames;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.common.IOUtils;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.xfer.FileSystemFile;
import net.schmizz.sshj.xfer.scp.SCPFileTransfer;

import java.io.File;
import java.util.Date;
import java.util.Set;
import java.util.concurrent.TimeUnit;

/**
 * Implementation for SCPAction.
 *
 * @author arul@flux.ly
 */
public class SCPActionImpl extends ActionImpl implements SCPAction {

  private static final String ACTION_VARIABLE = "SCP_ACTION_VARIABLE";

  public SCPActionImpl() {
    super(new FlowChartImpl(), "SCP Action");
  }

  public SCPActionImpl(FlowChartImpl fc, String name) {
    super(fc, name);
  }

  public Set<String> getHiddenVariableNames() {
    // The returned raw type is a Set of String.
    @SuppressWarnings("unchecked")
    Set<String> set = super.getHiddenVariableNames();
    set.add(ACTION_VARIABLE);
    return set;
  }

  @Override
  public void verify() throws EngineException {
    if (StringUtil.isNullOrEmpty(getHost())) {
      throw new EngineException("Expected \"Host\" to be non-null or non-empty, but it was null or empty.");
    } // if
    if (StringUtil.isNullOrEmpty(getUsername())) {
      throw new EngineException("Expected \"Username\" to be non-null or non-empty, but it was null or empty.");
    } // if
    if (StringUtil.isNullOrEmpty(getPrivateKey())) {
      final String base = System.getProperty("user.home") + File.separator + ".ssh" + File.separator;
      final String candidate1 = base + "id_rsa";
      final String candidate2 = base + "id_dsa";
      File file1 = new File(candidate1);
      File file2 = new File(candidate2);
      if (!file1.exists() && !file2.exists()) {
        throw new IllegalArgumentException("Expected \"Private key\" to be non-null or non-empty, but it was null or empty. No matching keys found in ~/.ssh.");
      }
    }
    if (!isRemoteExec()) {
      if (StringUtil.isNullOrEmpty(getLocalPath())) {
        throw new EngineException("Expected \"Local path\" to be non-null or non-empty, but it was null or empty.");
      } // if
      if (StringUtil.isNullOrEmpty(getRemotePath())) {
        throw new EngineException("Expected \"Remote path\" to be non-null or non-empty, but it was null or empty.");
      } // if
    }
  }

  @Override
  public void setHost(String host) {
    SCPActionVariable var = getVariable();
    var.setHost(host);
    putVariable(var);
  }

  public String getHost() {
    return getVariable().getHost();
  }

  @Override
  public void setCompression(boolean compression) {
    SCPActionVariable var = getVariable();
    var.setCompression(compression);
    putVariable(var);
  }

  public boolean getCompression() {
    return getVariable().isCompression();
  }

  @Override
  public void setRemoteCommand(String command) {
    SCPActionVariable var = getVariable();
    var.setRemoteCommand(command);
    putVariable(var);
  }

  public String getRemoteCommand() {
    return getVariable().getRemoteCommand();
  }

  @Override
  public void setPassword(String password) {
    SCPActionVariable var = getVariable();
    if (password != null) {
      var.setPassword(Password.makePassword(password));
    }
    putVariable(var);
  }

  public String getPassword() {
    Password password = getVariable().getPassword();
    if (password != null) {
      return password.getEncryptedPassword();
    }
    return null;
  }

  @Override
  public void setUsername(String username) {
    SCPActionVariable var = getVariable();
    var.setUsername(username);
    putVariable(var);
  }

  public String getUsername() {
    return getVariable().getUsername();
  }

  @Override
  public void setDownload(boolean download) {
    SCPActionVariable var = getVariable();
    var.setDownload(download);
    putVariable(var);
  }

  public boolean getDownload() {
    return getVariable().isDownload();
  }

  @Override
  public void setLocalPath(String localPath) {
    SCPActionVariable var = getVariable();
    var.setLocalPath(localPath);
    putVariable(var);
  }

  public String getLocalPath() {
    return getVariable().getLocalPath();
  }

  @Override
  public void setRemotePath(String remotePath) {
    SCPActionVariable var = getVariable();
    var.setRemotePath(remotePath);
    putVariable(var);
  }

  public String getRemotePath() {
    return getVariable().getRemotePath();
  }

  private boolean isRemoteExec() {
    return !StringUtil.isNullOrEmpty(getRemoteCommand());
  }

  @Override
  public void setPrivateKey(String privateKey) {
    SCPActionVariable var = getVariable();
    var.setPrivateKey(privateKey);
    putVariable(var);
  }

  public String getPrivateKey() {
    return getVariable().getPrivateKey();
  }

  @Override
  public Object execute(FlowContext flowContext) throws Exception {
    final Long timeout = calculateTimeout(getTimeoutExpression());
    final SSHClient ssh = new SSHClient();
//    ssh.addHostKeyVerifier(new PromiscuousVerifier());
    if (getCompression() && !isRemoteExec()) {
      ssh.useCompression();
    }
    if (timeout != 0) {
      ssh.setTimeout(timeout.intValue());
      ssh.setConnectTimeout(timeout.intValue());
    }
    ssh.loadKnownHosts();
    ssh.connect(getHost());
    try {
      if (!StringUtil.isNullOrEmpty(getPassword())) {
        ssh.authPassword(getUsername(), getPassword());
      } else {
        String privateKeyFile = getPrivateKey();
        if (StringUtil.isNullOrEmpty(privateKeyFile)) {
          ssh.authPublickey(getUsername());
        } else {
          ssh.authPublickey(getUsername(), privateKeyFile);
        }
      }
      if (isRemoteExec()) {
        final Session session = ssh.startSession();
        final SCPActionResult result = new SCPActionResult();
        try {
          final Session.Command cmd = session.exec(getRemoteCommand());
          cmd.join(5000, TimeUnit.MILLISECONDS);
          result.stdout = IOUtils.readFully(cmd.getInputStream()).toString();
          result.stderr = IOUtils.readFully(cmd.getErrorStream()).toString();
          result.result = cmd.getExitStatus();
        } finally {
          session.close();
        }
        flowContext.sendToAuditTrail("flux.audittrail.server.ExecutingActionEvent", getDetails(result, flowContext.getActionName()));
        return result;
      } else {
        final SCPFileTransfer scp = ssh.newSCPFileTransfer();
        if (getDownload()) {
          scp.download(getRemotePath(), new FileSystemFile(getLocalPath()));
        } else {
          scp.upload(new FileSystemFile(getLocalPath()), getRemotePath());
        }
      }
    } finally {
      ssh.disconnect();
    }
    return null;
  }

  public String getDetails(Object result, String actionName) {
    SCPActionResult scp = (SCPActionResult) result;
    StringBuilder sb = new StringBuilder();
    sb.append("<" + ElementNames.PROPERTY + ">");
    sb.append("<" + ElementNames.NAME + ">");
    sb.append("<![CDATA[");
    sb.append(AuditTrailConstants.RESULT);
    sb.append("]]>");
    sb.append("</" + ElementNames.NAME + ">");
    sb.append("<" + ElementNames.VALUE + ">");
    sb.append("<![CDATA[");
    sb.append(scp.result);
    sb.append("]]>");
    sb.append("</" + ElementNames.VALUE + ">");
    sb.append("</" + ElementNames.PROPERTY + ">");

    if (!StringUtil.isNullOrEmpty(scp.stdout)) {
      sb.append("<" + ElementNames.PROPERTY + ">");
      sb.append("<" + ElementNames.NAME + ">");
      sb.append("<![CDATA[");
      sb.append(AuditTrailConstants.STDOUT);
      sb.append("]]>");
      sb.append("</" + ElementNames.NAME + ">");
      sb.append("<" + ElementNames.VALUE + ">");
      sb.append("<![CDATA[");
      sb.append(scp.stdout);
      sb.append("]]>");
      sb.append("</" + ElementNames.VALUE + ">");
      sb.append("</" + ElementNames.PROPERTY + ">");
    }

    if (!StringUtil.isNullOrEmpty(scp.stderr)) {
      sb.append("<" + ElementNames.PROPERTY + ">");
      sb.append("<" + ElementNames.NAME + ">");
      sb.append("<![CDATA[");
      sb.append(AuditTrailConstants.STDERR);
      sb.append("]]>");
      sb.append("</" + ElementNames.NAME + ">");
      sb.append("<" + ElementNames.VALUE + ">");
      sb.append("<![CDATA[");
      sb.append(scp.stderr);
      sb.append("]]>");
      sb.append("</" + ElementNames.VALUE + ">");
      sb.append("</" + ElementNames.PROPERTY + ">");
    }

    return ElementNames.open(ElementNames.EVENT_MESSAGE_ENTRY)
        + ElementNames.open(ElementNames.EVENT_DESCRIPTION) + "Exiting " + actionName + " \"" + getName() + "\""
        + ElementNames.close(ElementNames.EVENT_DESCRIPTION)
        + ElementNames.open(ElementNames.MESSAGE_DETAIL)
        + ElementNames.open(ElementNames.DETAIL) + sb.toString()
        + ElementNames.close(ElementNames.DETAIL)
        + ElementNames.close(ElementNames.MESSAGE_DETAIL)
        + ElementNames.close(ElementNames.EVENT_MESSAGE_ENTRY);
  }

  private SCPActionVariable getVariable() {
    if (!getVariableManager().contains(ACTION_VARIABLE)) {
      getVariableManager().put(ACTION_VARIABLE, new SCPActionVariable());
    }
    return (SCPActionVariable) getVariableManager().get(ACTION_VARIABLE);
  }

  private void putVariable(SCPActionVariable variable) {
    getVariableManager().put(ACTION_VARIABLE, variable);
  }

  private Long calculateTimeout(String timeout) throws EngineException {
    Date date = Factory.makeInstance().makeEngineHelper().applyTimeExpression(timeout, null, null);
    return date.getTime() - System.currentTimeMillis();
  } // calculateTimeout
}
