package fluximpl;

import flux.EngineException;
import flux.FlowContext;
import flux.WebServiceTrigger;
import flux.dev.NotTriggeredException;
import flux.logging.Logger;
import fluximpl.logging.cxf.NoOpLogger;
import org.apache.cxf.common.logging.LogUtils;
import org.apache.cxf.endpoint.Server;
import org.apache.cxf.jaxrs.JAXRSServerFactoryBean;
import org.apache.cxf.jaxrs.lifecycle.SingletonResourceProvider;
import org.apache.cxf.jaxws.JaxWsProxyFactoryBean;
import org.apache.cxf.jaxws.JaxWsServerFactoryBean;
import org.apache.cxf.transport.http_jetty.JettyHTTPServerEngine;
import org.apache.cxf.transport.http_jetty.JettyHTTPServerEngineFactory;
import org.apache.cxf.ws.security.wss4j.WSS4JInInterceptor;
import org.apache.ws.security.WSConstants;
import org.apache.ws.security.handler.WSHandlerConstants;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Implementation for WebServiceTrigger.
 *
 * @author arul@flux.ly
 */
public class WebServiceTriggerImpl extends TriggerImpl implements WebServiceTrigger {

  public static final String TRIGGER_VARIABLE = "WEB_SERVICE_TRIGGER";
  private static final String HTTP = "http://";
  private static final String SEPARATOR = ":";
  private static final String SOAP_PATH = "/soap/";
  private static final String REST_PATH = "/rest/";

  public WebServiceTriggerImpl() {
    super(new FlowChartImpl(), "Web Service Trigger");
  }

  public WebServiceTriggerImpl(FlowChartImpl flowChart, String s) {
    super(flowChart, s);
  }

  public Set<String> getHiddenVariableNames() {
    // The returned raw type is a Set of String.
    @SuppressWarnings("unchecked")
    Set<String> set = super.getHiddenVariableNames();
    set.add(TRIGGER_VARIABLE);
    return set;
  }

  @Override
  public Date getNextPollingDate() {
    return null;
  }

  @Override
  public Object execute(FlowContext flowContext) throws Exception {
    verify();
    if (flowContext.getEngine().isSecured()) {
      if (StringUtil.isNullOrEmpty(getEngineUsername()) && StringUtil.isNullOrEmpty(getEnginePassword())) {
        throw new EngineException("Expected \"engine username\" and \"engine password\" to be non-null or non-empty for secured engine, but it was null or empty.");
      }
    }
    if (!checkIfWebServiceisRunning()) {
      setup(flowContext);
    }
    flowContext.getFlowChart().getVariableManager().put(FlowChartListenerImpl.LISTENER_EXPEDITED, Boolean.FALSE);
    throw new NotTriggeredException();
  }

  public Object expedite(FlowContext flowContext) throws Exception, NotTriggeredException {
    flowContext.getLogger().info("Expedite called for " + flowContext.getActionName());
    Map<String, String> variables = (Map) flowContext.getFlowChart().getVariableManager().get(FlowChartListenerImpl.LISTENER_VARIABLE);
    WebServiceTriggerResult result = new WebServiceTriggerResult();
    if (variables != null && variables.size() > 0) {
      result.variables = variables;
    }
    boolean triggerExpedited = (Boolean) flowContext.getFlowChart().getVariableManager().get(FlowChartListenerImpl.LISTENER_EXPEDITED);
    result.expedited = triggerExpedited;
    if (!triggerExpedited) {
      // Expedite called by other means, such as from OpsConsole or from Java API.
      disposeWebService();
    }
    return result;
  }

  /**
   * This delegation is required for pre 7.11 trigger expedite implementation.
   */
  public Object a(FlowContext flowContext) throws Exception, NotTriggeredException {
    return expedite(flowContext);
  }

  private void disposeWebService() {
    LogUtils.setLoggerClass(NoOpLogger.class);
    JaxWsProxyFactoryBean factory = new JaxWsProxyFactoryBean();
    factory.setServiceClass(FlowChartListener.class);
    try {
      factory.setAddress(getSoapEndpoint());
    } catch (Exception e) {
      // ignore
    }
    try {
      FlowChartListener listener = (FlowChartListener) factory.create();
      listener.dispose();
    } catch (Exception e) {
      // ignore
    }
  }

  private boolean checkIfWebServiceisRunning() {
    boolean running;
    LogUtils.setLoggerClass(NoOpLogger.class);
    JaxWsProxyFactoryBean factory = new JaxWsProxyFactoryBean();
    factory.setServiceClass(FlowChartListener.class);
    try {
      factory.setAddress(getSoapEndpoint());
    } catch (Exception e) {
      // ignore
    }
    try {
      FlowChartListener listener = (FlowChartListener) factory.create();
      running = listener.isRunning();
    } catch (Exception e) {
      running = false;
    }
    return running;
  }

  @Override
  public void verify() throws EngineException {
    if (getPort() < 0) {
      throw new EngineException("Expected \"port\" to be non-null or non-empty, but it was null or empty.");
    }
    if (StringUtil.isNullOrEmpty(getServiceName())) {
      throw new EngineException("Expected \"service name\" to be non-null or non-empty, but it was null or empty.");
    }
  }

  @Override
  public void setServiceName(String serviceName) {
    WebServiceTriggerVariable var = getVariable();
    String serviceNameWithoutSpaces = serviceName.replaceAll("\\s+", "");
    var.setServiceName(serviceNameWithoutSpaces);
    setVariable(var);
  }

  public String getServiceName() {
    return getVariable().getServiceName();
  }

  public int getPort() {
    return getVariable().getPort();
  }

  @Override
  public void setPort(int port) {
    WebServiceTriggerVariable var = getVariable();
    var.setPort(port);
    setVariable(var);
  }

  public String getPassword() {
    WebServiceTriggerVariable variable = getVariable();
    if (variable.getPassword() != null) {
      return variable.getPassword().getEncryptedPassword();
    }
    return null;
  }

  public String getUsername() {
    return getVariable().getUsername();
  }

  public String getEnginePassword() {
    WebServiceTriggerVariable variable = getVariable();
    if (variable.getEnginePassword() != null) {
      return variable.getEnginePassword().getEncryptedPassword();
    }
    return null;
  }

  public String getEngineUsername() {
    return getVariable().getEngineUsername();
  }

  @Override
  public void setPassword(String password) {
    WebServiceTriggerVariable var = getVariable();
    var.setPassword(Password.makePassword(password));
    setVariable(var);
  }

  @Override
  public void setUsername(String username) {
    WebServiceTriggerVariable var = getVariable();
    var.setUsername(username);
    setVariable(var);
  }

  @Override
  public void setEnginePassword(String enginePassword) {
    WebServiceTriggerVariable var = getVariable();
    var.setEnginePassword(Password.makePassword(enginePassword));
    setVariable(var);
  }

  @Override
  public void setEngineUsername(String engineUsername) {
    WebServiceTriggerVariable var = getVariable();
    var.setEngineUsername(engineUsername);
    setVariable(var);
  }

  private WebServiceTriggerVariable getVariable() {
    if (!getVariableManager().contains(TRIGGER_VARIABLE)) {
      getVariableManager().put(TRIGGER_VARIABLE, new WebServiceTriggerVariable());
    }
    return (WebServiceTriggerVariable) getVariableManager().get(TRIGGER_VARIABLE);
  }

  private void setVariable(WebServiceTriggerVariable variable) {
    getVariableManager().put(TRIGGER_VARIABLE, variable);
  }

  private String constructEndpoint() throws Exception {
    String host;
    try {
      host = InetAddress.getLocalHost().getHostName();
    } catch (UnknownHostException e) {
      try {
        host = InetAddress.getLocalHost().getHostAddress();
      } catch (UnknownHostException e1) {
        throw new EngineException(e);
      }
    }
    return HTTP + host + SEPARATOR + getPort();
  }

  private String getRestEndpoint() throws Exception {
    String address = constructEndpoint();
    return address + REST_PATH + getServiceName();
  }

  private String getSoapEndpoint() throws Exception {
    String address = constructEndpoint();
    return address + SOAP_PATH + getServiceName();
  }

  @SuppressWarnings("unchecked")
  private void setup(FlowContext flowContext) throws Exception {
    Logger logger = flowContext.getLogger();
    LogUtils.setLoggerClass(NoOpLogger.class);
    String engineUsername = null;
    String enginePassword = null;
    if (flowContext.getEngine().isSecured()) {
      engineUsername = getVariable().getEngineUsername();
      enginePassword = getVariable().getEnginePassword().getClearTextPassword();
    }
    FlowChartListenerImpl flowChartListener = new FlowChartListenerImpl(flowContext, engineUsername, enginePassword);
    JettyHTTPServerEngineFactory factory = new JettyHTTPServerEngineFactory();
    JettyHTTPServerEngine engine = factory.createJettyHTTPServerEngine(getPort(), "http");
    String restEndpoint = getRestEndpoint();
    String soapEndpoint = getSoapEndpoint();

    //Setup SOAP Server
    JaxWsServerFactoryBean soapFactory = new JaxWsServerFactoryBean();
    soapFactory.setServiceClass(FlowChartListener.class);
    soapFactory.setServiceBean(flowChartListener);
    soapFactory.setAddress(soapEndpoint);
    if (isSecured()) {
      Map<String, Object> inProps = new HashMap<String, Object>();
      inProps.put(WSHandlerConstants.ACTION, WSHandlerConstants.USERNAME_TOKEN);
      inProps.put(WSHandlerConstants.PASSWORD_TYPE, WSConstants.PW_DIGEST);
      String password = getVariable().getPassword().getClearTextPassword();
      inProps.put(WSHandlerConstants.PW_CALLBACK_REF, new ServerPasswordCallbackHandler(getUsername(), password));
      WSS4JInInterceptor wssIn = new WSS4JInInterceptor(inProps);
      soapFactory.getInInterceptors().add(wssIn);
    }
    Server soapServer = soapFactory.create();
    logger.info("SOAP Server listening @ " + soapServer.getEndpoint().getEndpointInfo().getAddress());

    //Setup REST Server
    JAXRSServerFactoryBean restFactory = new JAXRSServerFactoryBean();
    restFactory.setResourceClasses(FlowChartListener.class);
    restFactory.setResourceProvider(FlowChartListener.class, new SingletonResourceProvider(flowChartListener));
    restFactory.setAddress(restEndpoint);
    if (isSecured()) {
      String password = getVariable().getPassword().getClearTextPassword();
      restFactory.setProvider(new AuthenticationHandler(getUsername(), password));
    }

    Server restServer = restFactory.create();
    logger.info("REST Server listening @ " + restServer.getEndpoint().getEndpointInfo().getAddress());
    flowChartListener.setRestServer(restServer);
    flowChartListener.setSoapServer(soapServer);
    flowChartListener.setEngine(engine);
  }

  private boolean isSecured() {
    return !StringUtil.isNullOrEmpty(getUsername()) && (!StringUtil.isNullOrEmpty(getUsername()));
  }
}